#!/usr/bin/env python3
import sys
assert sys.version_info[0] >= 3, "Python 3 required."

import subprocess
import os
import shutil

footer_banks = 0x7C # graphics and code data begins here

music_list = [
    "music/bun%020-%020new years.nsf",               # 00
    "music/DJ Dimeback%20-%20Adrenaline Rush.nsf",   # 01
    "music/dusthillguy%20-%20.nsf",                  # 02
    "music/gyrouh%20-%20Benny the Seal.nsf",         # 03
    "music/keffie%20-%20MOLOCH.nsf",                 # 04
    "music/maak%20-%20emotional magic.nsf",          # 05
    "music/MXSP%20-%20Hi-tech.nsf",                  # 06
    "music/OffGao%20-%20Otenba Tomboy.nsf",          # 07
    "music/rolemusic%20-%20Yokushun.nsf",            # 08
    "music/RushJet1%20-%20The Grand Imbecile.nsf",   # 09
    "music/Tom Miller%20-%20Ancient Melody.nsf",     # 0A
]

music_order = [
    # 00    01    02    03    04    05    06    07    08    09    0A
    0x03, 0x05, 0x06, 0x0A, 0x09, 0x00, 0x07, 0x02, 0x04, 0x08, 0x01
]

music_time = [
    [ 3,35+0], # 00 loops
    [ 2,45+2], # 01
    [ 1,58+2], # 02
    [ 2,31+2], # 03
    [ 1,43+2], # 04
    [ 2,56+0], # 05 loops
    [ 3,48+2], # 06
    [ 3,26+2], # 07
    [ 3,43+2], # 08 (loops after pause)
    [ 1,46+0], # 09 loops
    [ 2,15+2], # 0A
]

music_pal = [
    # 01 02 03 04 05 06 07 08 09 0A
    1, 1, 0, 1, 0, 0, 0, 1, 1, 1, 0,
]

music_loop = [
    # 01 02 03 04 05 06 07 08 09 0A
    1, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0,
]

# the following banks need to be patched
disassemble_list = {
    "00_00" : 0x8000, # $500-7FF RAM init
    "03_07" : 0xF000, # vectors
    "0A_07" : 0xF000, # vectors
    "0A_0B" : 0xF000, # vectors
    "0A_0F" : 0xF000, # vectors
    "01_01" : 0x9000, # banking code
    "08_00" : 0x8000, # banking code
    "0A_00" : 0x8000, # banking code
    "0A_01" : 0x9000, # banking code
}

bank_override = {
    (0x01,0x0C) : 0xFF, # strangely assigned unused banks out-of-file?
    (0x08,0x07) : 0xFF, # this bank is unused but listed in the setup bank list
}

init_list = "track_init_addr:\n"
play_list = "track_play_addr:\n"
bank_list = "track_bank_offset: ; bank location of NSF in ROM\n"
bstr_list = "track_bank_start: ; starting banks for NSF\n"
bstr_list += ";     8000 9000 A000 B000 C000 D000 E000 F000\n"
pala_list = "track_pal_adjust: ; double every 5th frame if in PAL\n"
pala_list += ";     " + "".join([("%02X " % x) for x in range(0,len(music_pal))]) + "end\n"
pala_list += ".byte "
loop_list = "track_loop: ; whether track loops\n"
loop_list += ";     " + "".join([("%02X " % x) for x in range(0,len(music_loop))]) + "end\n"
loop_list += ".byte "
time_list = "track_time: ; time to play each track in seconds\n"
ordr_list = "track_order:\n"
ordr_list += ".byte "
copy_list = "REM track data\n"

bank_offset = 0x02

def prepend_org(filename,org):
    f = open(filename,"r+")
    text = f.read()
    f.seek(0)
    f.write((".org $%04X\n" % org) + text)
    f.close()

def disassemble(name, address):
    file_in = "src_bin\\" + name + ".bin"
    file_out = "src_bin\\" + name + ".s"
    cline = "cc65\\bin\\da65 -o %s -v --comments 4 -g -S $%04X %s" % (file_out,address,file_in)
    print (cline)
    subprocess.call(cline)
    prepend_org(file_out, address)

def unpack_nsf(name,index):
    print ("NSF %02X: %s" % (index,name));
    if not os.path.exists(name):
        print("File not found!")
        return
    nsf = open(m,"rb").read()
    # check for bankswitching
    bankswitching = False
    for b in range(0,8):
        if nsf[0x070 + b] != 0:
            bankswitching = True
    load_addr = nsf[0x008] + (nsf[0x009] << 8)
    init_addr = nsf[0x00A] + (nsf[0x00B] << 8)
    play_addr = nsf[0x00C] + (nsf[0x00D] << 8)
    print ("LOAD: $%04X" % load_addr)
    print ("INIT: $%04X" % init_addr)
    print ("PLAY: $%04X" % play_addr)
    global init_list
    global play_list
    global bank_list
    global bstr_list
    global pala_list
    global loop_list
    global time_list
    global copy_list
    global bank_offset
    init_list += ".word $%04X ; %02X %s\n" % (init_addr, index, name)
    play_list += ".word $%04X ; %02X %s\n" % (play_addr, index, name)
    bank_list += ".byte $%02X ; %02X %s\n" % (bank_offset, index, name)
    bstr_list += ".byte "
    pala_list += "%d, " % (music_pal[index])
    loop_list += "%d, " % (music_loop[index])
    time_list += ".word (%2d*60)+%2d ; %02X %s\n" % (music_time[index][0],music_time[index][1],index, name)
    if bankswitching:
        load_addr = 0x8000 | (load_addr & 0x0FFF)
    # build image
    if (load_addr < 0x8000):
        print("LOAD address too low!")
        return
    high_addr = load_addr + (len(nsf) - 128)
    bank_count = 8
    while high_addr > (0x8000 + (bank_count * 0x1000)):
        bank_count += 1
    if bankswitching:
        for b in range(0,8):
            sb = nsf[0x070 + b]
            if (index,sb) in bank_override:
                sb = bank_override[(index,sb)]
            bstr_list += " $%02X" % sb
            if (b != 7):
                bstr_list += ","
    b = bytearray([0] * (1024 * 1024))
    for a in range(128,len(nsf)):
        nb = nsf[a]
        wa = (load_addr + a) - (0x8000 + 128)
        b[wa] = nb
    # dump chunks to file
    chunk = 0
    for a in range(0,bank_count):
        ra = (a * 0x1000) # image start address
        na = ra + 0x8000 # bank start address
        l = 0x1000 # length
        fn = "src_bin\\%02X_%02X.bin" % (index,chunk)
        if (na + l) > load_addr and na < high_addr:
            bo = b[ra:(ra + l)]
            print("Writing "+fn+"...")
            open(fn,"wb").write(bo)
            if not bankswitching:
                bstr_list += " $%02X" % chunk
            if ("%02X_%02X" % (index,chunk)) in disassemble_list:
                copy_list += "REM "
            copy_list += "copy /b %s  temp\\%02X.bin\n" % (fn,bank_offset)
            chunk += 1
            bank_offset += 1
        elif not bankswitching:
            bstr_list += " $%02X" % 0xFF
        if not bankswitching and (a != 7):
            bstr_list += ","
    if bankswitching:
        bstr_list += " ; BANKD "
    else:
        bstr_list += " ; FIXED "
    bstr_list += "%02X %s\n" % (index, name)

#
# main
#

if os.path.exists("src_bin"):
    shutil.rmtree("src_bin")
os.mkdir("src_bin")

for i in range(0,len(music_list)):
    m = music_list[i]
    if (len(m) == 0):
        continue
    unpack_nsf(m,i)
    print("Done unpacking.")
    print()
    
for d in disassemble_list:
    org = disassemble_list[d]
    disassemble(d,org)
    print ("disassembled: " + d + " at $%04X" % org)

pala_list += "0\n"
loop_list += "0\n"

for i in range(0,len(music_order)):
    ordr_list += "$%02X, " % music_order[i]
ordr_list += "0\n"

while bank_offset < footer_banks:
    copy_list += "copy /b empty4k.bin  temp\\%02X.bin\n" % bank_offset
    bank_offset += 1

s = "\n" + \
    "; tables generated by disassemble_music.py\n" + \
    "\n" + \
    init_list + "\n" + \
    play_list + "\n" + \
    bank_list + "\n" + \
    bstr_list + "\n" + \
    pala_list + "\n" + \
    loop_list + "\n" + \
    time_list + "\n" + \
    ordr_list + "\n" + \
    "; end of generated tables\n"

print(s)
open("music_tables.inc","wt").write(s)
    
print(copy_list)
open("copy_list.bat","wt").write(copy_list)
